/*
 *  =======================================================================
 *  LineSegment2D.java: A line segment is defined by the (x,y) coordinates
 *                      of its two end points. 
 * 
 *  Written By: Mark Austin                                      April 2019 
 *  =======================================================================
 */

package geometry;

import java.lang.Math;

public class LineSegment2D {
   private Node2D   n1, n2;   // nodal points defining the line segment ...
   private BoundingBox box;   // bounding box for line segment ...
   private String name = null;

   public static final double EPSILON = 0.0000001;

   // -------------------------------------------------------
   // Constructor methods ...
   // -------------------------------------------------------

   public LineSegment2D() {
      box = new BoundingBox();
   }

   public LineSegment2D( Node2D n1, Node2D n2 ) {
      this.box = new BoundingBox( n1, n2 );
      this.n1 = n1;
      this.n2 = n2;
   }

   public LineSegment2D( double dX1, double dY1, double dX2, double dY2 ) {
      box = new BoundingBox( dX1, dX2, dY1, dY2 );
      n1  = new Node2D( dX1, dY1 );
      n2  = new Node2D( dX2, dY2 );
   }

   // -------------------------------------------------------
   // Get end points n1 and s2.
   // -------------------------------------------------------

   public Node2D getNode1() {
      return n1;
   }  

   public Node2D getNode2() {
      return n2;
   }

   // -------------------------------------------------------
   // Set/get line segment name ...
   // -------------------------------------------------------

   public void setName( String name ) {
      this.name = name;
   }

   public String getName() {
      return name;
   }

   // -------------------------------------------------------
   // Retrieve bounding box ...
   // -------------------------------------------------------

   public BoundingBox getBoundingBox() {
      return box;
   }

   // -------------------------------------------------------
   // Compute length of line segment.
   // -------------------------------------------------------

   public double getLength() {
      double dLength;

      dLength = (n1.getX() - n2.getX())*(n1.getX() - n2.getX()) +
                (n1.getY() - n2.getY())*(n1.getY() - n2.getY());

      return Math.sqrt(dLength);
   }

   // ---------------------------------------------------------
   // Checks to see if a point is "on a line segment"?
   // If yes,
   //    return true.
   // Otherwise,
   //    return false.
   // ---------------------------------------------------------

   public boolean isPointOnLine( Node2D point ) {

      // Create "direction" line segment ...

      Node2D t1 = new Node2D( 0.0, 0.0 );

      Vector2D direction = n2.sub(n1);
      Node2D t2 = new Node2D( direction.getX(),  direction.getY() );

      LineSegment2D aTmp = new LineSegment2D( t1, t2 );

      // create temporary point ...

      Vector2D bTmp = new Vector2D( point.getX() - n1.getX(),
                                    point.getY() - n1.getY() );

      double r = t2.crossProduct( bTmp );

      return Math.abs(r) < EPSILON;
   }

   // ----------------------------------------------------------
   // Check to see if a point is to the right of a line segment.
   // ----------------------------------------------------------

   public boolean isPointRightOfLine( Node2D point ) {

      // Create "direction" line segment ...

      Node2D t1 = new Node2D( 0.0, 0.0 );

      Vector2D direction = n2.sub(n1);
      Node2D t2 = new Node2D( direction.getX(),  direction.getY() );

      LineSegment2D aTmp = new LineSegment2D( t1, t2 );

      // create temporary point ...

      Vector2D bTmp = new Vector2D( point.getX() - n1.getX(),
                                    point.getY() - n1.getY() );

      double r = t2.crossProduct( bTmp );
      return r < 0;
   }

   // ----------------------------------------------------------
   // If first line segment touches or crosses line second,
   //    return true,
   // Otherwise
   //    return false.
   // ----------------------------------------------------------

   public boolean lineSegmentTouchesOrCrossesLine( LineSegment2D test ) {
      boolean part01 = this.isPointOnLine( test.getNode1() );
      boolean part02 = this.isPointOnLine( test.getNode2() );
      boolean part03 = this.isPointRightOfLine( test.getNode1() );
      boolean part04 = this.isPointRightOfLine( test.getNode2() );

      boolean result = part01 || part02 || part03^part04;
      return result;
   }

   // -------------------------------------------------------
   // Check to see if two line segments intersect ...
   // -------------------------------------------------------

   public boolean intersect( LineSegment2D test ) {

      // Check to see if bounding boxes intersect ...

      BoundingBox box1 = this.getBoundingBox();
      BoundingBox box2 = test.getBoundingBox();

      boolean part1 = box1.intersect( box2 );

      // Check to see if line segments touch or cross ...

      boolean part2 = this.lineSegmentTouchesOrCrossesLine( test );
      boolean part3 = test.lineSegmentTouchesOrCrossesLine( this );

      boolean result = part1 && part2 && part3;
      return result;
   }

   // -------------------------------------------------------
   // Compute min distance from line segment to point ...
   // -------------------------------------------------------

   public double distanceToPoint( Node2D pt1 ) {

      // Get vector direction along line segment ...

      Vector2D direction = n2.sub(n1);

      // Compute intercept point ....

      double rhs = pt1.sub(n1).dotProduct ( direction );
      double lhs =   direction.dotProduct ( direction );
      double t = rhs/lhs;

      // If intercept lies in "end zones" compute Euclidean distance
      // to end points of line segments....

      if ( 0.0 < t && t < 1.0 )
           return pt1.sub( n1.add( direction.scale(t)) ).length();
      else 
           return Math.min ( pt1.sub( n1 ).length(), pt1.sub( n2 ).length() );
   }

   // ----------------------------------------------------------
   // Equation of line segment: P = r*v1 + (1-r)*v2, 0 <= r <= 1
   // ----------------------------------------------------------

   public String getEquation(){
      StringBuffer buf = new StringBuffer();

      Vector2D direction = n2.sub(n1);

      buf.append( String.format( "equation: (x,y) = (%f, %f) + r*(%f, %f), 0 < r < 1",
                                 n1.getX(), n1.getY(), direction.getX(), direction.getY() ));

      return buf.toString();
   } 

   // -------------------------------------------------------
   // Create string representation of line segment ...
   // -------------------------------------------------------

   public String toString() {
      StringBuffer buf = new StringBuffer();
   
      if ( name == null )
         buf.append("Line Segment \n");
      else
         buf.append( String.format( "Line Segment: %s ...\n", name ));

      buf.append(String.format("--- node 1: (x,y) = (%.3f, %.3f) ...\n", n1.getX(), n1.getY() ));
      buf.append(String.format("--- node 2: (x,y) = (%.3f, %.3f) ...\n", n2.getX(), n2.getY() ));
      buf.append(String.format("--- length = %8.2f ...\n", getLength() ));
      buf.append(String.format("--- %s ...\n", getEquation() ));
      buf.append(String.format("--- %s ...\n", box.toString() ));

      return buf.toString();
   }
}
