/* 
 *  ========================================================================
 *  DemoTimeSeriesChart02.java: Create time series plot by reading data from
 *                              csv file.
 * 
 *  Written by: Mark Austin                                    November 2021 
 *  ========================================================================
 */ 

package jfreechart;

import au.com.bytecode.opencsv.CSVReader;
import au.com.bytecode.opencsv.CSVWriter;

import java.io.FileReader;
import java.io.IOException;
import java.io.StringWriter;

import java.util.List;
import java.util.Date;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.block.BlockBorder;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.title.TextTitle;
import org.jfree.chart.ui.RectangleInsets;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.jfree.data.time.Day;
import org.jfree.data.time.Month;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;

import java.text.SimpleDateFormat;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.EventQueue;
import java.awt.Font;

public class DemoTimeSeriesChart02 extends JFrame {

    // Constructor meethods ...

    public DemoTimeSeriesChart02() {
       try {
          initUI();
       } catch(IOException e) {}
    }

    public DemoTimeSeriesChart02( String name ) {
       super (name);
       try {
          initUI();
       } catch(IOException e) {}

    }

    // Initialize datasets ...

    private void initUI() throws IOException {
        XYDataset dataset = createDataset();
        JFreeChart chart  = createChart(dataset);

        ChartPanel chartPanel = new ChartPanel(chart);
        chartPanel.setBorder(BorderFactory.createEmptyBorder(15, 15, 15, 15));
        chartPanel.setBackground(Color.white);
        
        add(chartPanel);

        pack();
        setLocationRelativeTo(null);
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    }

    // ================================================================
    // Create dataset from data/stocks.csv
    // ================================================================

    private XYDataset createDataset() throws IOException {

        // Create placeholders for stock time series ....

        TimeSeries msft = new TimeSeries("MSFT stock");
        TimeSeries aapl = new TimeSeries("AAPL stock");
        TimeSeries  ibm = new TimeSeries("IBM stock");
        TimeSeries goog = new TimeSeries("GOOG stock");

        // Create CSVReader and skip over header ...

        CSVReader reader = new CSVReader( new FileReader("data/stocks.csv") );
        String header [] = reader.readNext();

        // Loop over individual stock indicies ...

        String [] nextLine;
        while ((nextLine = reader.readNext()) != null) {
            String stock = nextLine[0];
            String date  = nextLine[1];
            Double stockPrice = Double.parseDouble( nextLine[2] );

            // Create date object from string input ...

            Date dateItem = new Date( date );
           
            // Process individual stock profiles ...

            switch ( stock ) {
               case "MSFT":
                    System.out.printf("--- Processing MSFT: %.2f ...\n", stockPrice );
                    msft.add( new Month( dateItem ), stockPrice );
                    break;
               case "AAPL":
                    System.out.printf("--- Processing AAPL: %.2f ...\n", stockPrice );
                    aapl.add( new Month( dateItem ), stockPrice );
                    break;
               case "IBM":
                    System.out.printf("--- Processing IBM: %.2f ...\n", stockPrice );
                    ibm.add( new Month( dateItem ), stockPrice );
                    break;
               case "GOOG":
                    System.out.printf("--- Processing GOOG: %.2f ...\n", stockPrice );
                    goog.add( new Month( dateItem ), stockPrice );
                    break;
               default:
                    break;
            }
        }

        // Assemble collection of x-y lines ...

        TimeSeriesCollection dataset = new TimeSeriesCollection();
        dataset.addSeries(msft);
        dataset.addSeries(aapl);
        dataset.addSeries(goog);
        dataset.addSeries(ibm);

        return dataset;
    }

    private JFreeChart createChart(final XYDataset dataset) {

        JFreeChart chart = ChartFactory.createTimeSeriesChart(
            "Stock Price vs Time", // title
            "Date",                // x-axis label
            "Stock Price (USD)",   // y-axis label
            dataset,               // data
            true,                  // create legend?
            true,                  // generate tooltips?
            false                  // generate URLs?
        );

        chart.setBackgroundPaint(Color.white);

        XYPlot plot = (XYPlot) chart.getPlot();
        plot.setBackgroundPaint(Color.white);
        plot.setDomainGridlinePaint(Color.black);
        plot.setRangeGridlinePaint(Color.black);
        plot.setAxisOffset( new RectangleInsets(5.0, 5.0, 5.0, 5.0) );
        plot.setDomainCrosshairVisible(true);
        plot.setRangeCrosshairVisible(true);
       
        XYItemRenderer r = plot.getRenderer();
        if (r instanceof XYLineAndShapeRenderer) {
            XYLineAndShapeRenderer renderer = (XYLineAndShapeRenderer) r;
            renderer.setSeriesPaint(0, Color.RED);
            renderer.setSeriesStroke(0, new BasicStroke(2.0f));
            renderer.setSeriesPaint(1, Color.BLUE);
            renderer.setSeriesStroke(1, new BasicStroke(2.0f));
        }

        DateAxis axis = (DateAxis) plot.getDomainAxis();
        axis.setDateFormatOverride(new SimpleDateFormat("MMM-yyyy"));

        chart.getLegend().setFrame(BlockBorder.NONE);
        chart.setTitle(new TextTitle("Time-History of Stock Prices", new Font("Serif", Font.BOLD, 18)));

        return chart;
    }

    public static void main(String[] args) {
        EventQueue.invokeLater(() -> {
            DemoTimeSeriesChart02 ex = new DemoTimeSeriesChart02("Time-History of Stock Prices");
            ex.setVisible(true);
        });
    }
}
